@@ -3,23 +3,20 @@ import cgi
 import logging
 import os
 import re
-import sys
+from urllib.parse import unquote_plus
 
 import cachecontrol
 import cryptography.hazmat.backends
 import jwt
 import requests
+import requests.utils
 from cryptography.hazmat.primitives import serialization
-from requests.exceptions import RequestException
+from requests.exceptions import RequestException, ConnectionError
 
 from atlassian_jwt_auth.exceptions import (KeyIdentifierException,
                                            PublicKeyRetrieverException,
                                            PrivateKeyRetrieverException)
 
-if sys.version_info[0] >= 3:
-    from urllib.parse import unquote_plus
-else:
-    from urllib import unquote_plus
 
 PEM_FILE_TYPE = 'application/x-pem-file'
 
@@ -38,7 +35,7 @@ class KeyIdentifier(object):
 
 def validate_key_identifier(identifier):
     """ returns a validated key identifier. """
-    regex = re.compile('^[\w.\-\+/]*$')
+    regex = re.compile(r'^[\w.\-\+/]*$')
     _error_msg = 'Invalid key identifier %s' % identifier
     if not identifier:
         raise KeyIdentifierException(_error_msg)
@@ -72,6 +69,9 @@ class HTTPSPublicKeyRetriever(BasePublicKeyRetriever):
     """ This class retrieves public key from a https location based upon the
          given key id.
     """
+    # Use a static requests session, reused/shared by all instances of
+    # HTTPSPublicKeyRetriever:
+    _class_session = None
 
     def __init__(self, base_url):
         if base_url is None or not base_url.startswith('https://'):
@@ -81,17 +81,21 @@ class HTTPSPublicKeyRetriever(BasePublicKeyRetriever):
             base_url += '/'
         self.base_url = base_url
         self._session = self._get_session()
+        self._proxies = requests.utils.get_environ_proxies(self.base_url)
 
     def _get_session(self):
-        session = requests.Session()
-        session.mount('https://', cachecontrol.CacheControlAdapter())
-        return session
+        if HTTPSPublicKeyRetriever._class_session is None:
+            session = cachecontrol.CacheControl(requests.Session())
+            session.trust_env = False
+            HTTPSPublicKeyRetriever._class_session = session
+        return HTTPSPublicKeyRetriever._class_session
 
     def retrieve(self, key_identifier, **requests_kwargs):
         """ returns the public key for given key_identifier. """
         if not isinstance(key_identifier, KeyIdentifier):
             key_identifier = KeyIdentifier(key_identifier)
-
+        if self._proxies and 'proxies' not in requests_kwargs:
+            requests_kwargs['proxies'] = self._proxies
         url = self.base_url + key_identifier.key_id
         try:
             return self._retrieve(url, requests_kwargs)
@@ -132,15 +136,29 @@ class HTTPSMultiRepositoryPublicKeyRetriever(BasePublicKeyRetriever):
         return [HTTPSPublicKeyRetriever(url) for url
                 in key_repository_urls]
 
+    def handle_retrieval_exception(self, retriever, exception):
+        """ Handles working with exceptions encountered during key
+            retrieval.
+        """
+        if isinstance(exception, PublicKeyRetrieverException):
+            original_exception = getattr(
+                exception, 'original_exception', None)
+            if isinstance(original_exception, ConnectionError):
+                return
+            if exception.status_code is None or exception.status_code < 500:
+                raise
+
     def retrieve(self, key_identifier, **requests_kwargs):
         for retriever in self._retrievers:
             try:
                 return retriever.retrieve(key_identifier, **requests_kwargs)
-            except RequestException as e:
+            except (RequestException, PublicKeyRetrieverException) as e:
+                self.handle_retrieval_exception(retriever, e)
                 logger = logging.getLogger(__name__)
-                logger.warn('Unable to retrieve public key from store',
-                            extra={'underlying_error': str(e),
-                                   'key repository': retriever.base_url})
+                logger.warning(
+                    'Unable to retrieve public key from store',
+                    extra={'underlying_error': str(e),
+                           'key repository': retriever.base_url})
         raise PublicKeyRetrieverException(
             'Cannot load key from key repositories')
 
